#include "skein/orbit.h"

#include "astro/stateVectorIndices.hpp"
#include "astro/orbitalElementConversions.hpp"
#include "astro/twoBodyMethods.hpp"

#include <glm/gtc/matrix_transform.hpp>

Orbit::Orbit()
{
    _keplerianElements.resize(6);
}

double Orbit::getSemiMajorAxis() const
{
    return _keplerianElements[astro::semiMajorAxisIndex];
}
void Orbit::setSemiMajorAxis(double semiMajorAxis)
{
    _keplerianElements[astro::semiMajorAxisIndex] = semiMajorAxis;
}

double Orbit::getEccentricity() const
{
    return _keplerianElements[astro::eccentricityIndex];
}
void Orbit::setEccentricity(double eccentricity)
{
    _keplerianElements[astro::eccentricityIndex] = eccentricity;
}

double Orbit::getInclination() const
{
    return _keplerianElements[astro::inclinationIndex];
}
void Orbit::setInclination(double inclination)
{
    _keplerianElements[astro::inclinationIndex] = inclination;
}

double Orbit::getArgumentOfPeriapsis() const
{
    return _keplerianElements[astro::argumentOfPeriapsisIndex];
}
void Orbit::setArgumentOfPeriapsis(double argumentOfPeriapsis)
{
    _keplerianElements[astro::argumentOfPeriapsisIndex] = argumentOfPeriapsis;
}

double Orbit::getLongitudeOfAscendingNode() const
{
    return _keplerianElements[astro::longitudeOfAscendingNodeIndex];
}
void Orbit::setLongitudeOfAscendingNode(double longitudeOfAscendingNode)
{
    _keplerianElements[astro::longitudeOfAscendingNodeIndex] = longitudeOfAscendingNode;
}

//glm::mat4 Orbit::getLookAlongMatrix(const double meanAnomaly) const
//{
//    glm::vec3 position = getPosition(meanAnomaly);
//
//    // get the tangent of the orbital ellipse
//    glm::vec3 tan = getTangent(meanAnomaly);
//    // we want to point along the orbit
//    glm::vec3 target = position + tan;
//    // TODO: this is not 'up' with respect to the orbited body!
//    // 'up' is just the normalized position vector because the orbit is centred at the origin
//    glm::vec3 up = glm::normalize(position);
//    // easy peasy lookAt matrix
//    glm::mat4 look = glm::lookAt(position, target, up);
//
//    // invert the lookat matrix because it's meant for cameras, cameras work backwards and
//    // we are not a camera
//    glm::mat4 lookAlong = glm::inverse(look);
//
//    return lookAlong;
//}

const double Orbit::getEccentricAnomaly(const double meanAnomaly) const
{
    const double eccentricity = _keplerianElements[astro::eccentricityIndex];
    double eccentricAnomaly = astro::convertEllipticalMeanAnomalyToEccentricAnomalyBS(
        eccentricity,
        meanAnomaly,
        200);

    return eccentricAnomaly;
}

glm::dvec3 Orbit::getPositionFromMeanAnomaly(const double meanAnomaly, double gravitationalParameter) const
{
    Vector6 cartesian = getCartesianCoordinates(meanAnomaly, gravitationalParameter);
    return glm::dvec3(
        cartesian[astro::xPositionIndex],
        cartesian[astro::yPositionIndex],
        cartesian[astro::zPositionIndex]);
}

glm::dvec3 Orbit::getPosition(double gravitationalParameter, double time) const
{
    return getPositionFromMeanAnomaly(getMeanAnomaly(gravitationalParameter, time), gravitationalParameter);
}

glm::dvec3 Orbit::getVelocityFromMeanAnomaly(const double meanAnomaly, double gravitationalParameter) const
{
    Vector6 cartesian = getCartesianCoordinates(meanAnomaly, gravitationalParameter);
    return glm::dvec3(
        cartesian[astro::xVelocityIndex],
        cartesian[astro::yVelocityIndex],
        cartesian[astro::zVelocityIndex]);
}

glm::dvec3 Orbit::getVelocity(double gravitationalParameter, double time) const
{
    return getVelocityFromMeanAnomaly(getMeanAnomaly(gravitationalParameter, time), gravitationalParameter);
}

const double Orbit::getMeanAnomaly(const double gravitationalParameter, const double time) const
{
    double meanMotion = astro::computeKeplerMeanMotion(getSemiMajorAxis(), gravitationalParameter);
    return meanMotion * time;
}

const double Orbit::getTrueAnomaly(const double meanAnomaly) const
{
    const double eccentricAnomaly = getEccentricAnomaly(meanAnomaly);
    const double eccentricity = _keplerianElements[astro::eccentricityIndex];
    return astro::convertEccentricAnomalyToTrueAnomaly(
        eccentricAnomaly,
        eccentricity);
}

const Vector6 Orbit::getCartesianCoordinates(const double meanAnomaly, const double gravitationalParameter) const
{
    Vector6 kepler(_keplerianElements);
    kepler[astro::trueAnomalyIndex] = getTrueAnomaly(meanAnomaly);
    return astro::convertKeplerianToCartesianElements(kepler, gravitationalParameter);
}

//glm::dvec3 Orbit::getTangent(const double meanAnomaly) const
//{
//    double epsilon = 0.01;
//    glm::dvec3 ahead = getPosition(meanAnomaly + epsilon);
//    glm::dvec3 behind = getPosition(meanAnomaly - epsilon);
//    return glm::normalize(ahead - behind);
//}

