#include "skein/particlemap.h"
#include "skein/particle.h"
#include "astro/orbitalElementConversions.hpp"

const Particle& ParticleMap::getParticle(const std::string& id) const
{
    return _particles.at(id);
}

const Particle& ParticleMap::getParent(const std::string& childId) const
{
    const std::string parentId = _relationships.at(childId);
    return _particles.at(parentId);
}

const Orbit& ParticleMap::getOrbit(const std::string& id) const
{
    return _orbits.at(id);
}

glm::dvec3 ParticleMap::getParticleLocalPosition(const std::string& id, double time) const
{
    if (_orbits.find(id) == _orbits.end())
        return {0,0,0};

    const Particle* particle = &(_particles.at(id));
    glm::dvec3 pos(0,0,0);

    //const std::string& id = particle->getId();
    const std::string& parentId = _relationships.at(id);
    const Particle& parent = _particles.at(parentId);
    const double gravitationalParameter = parent.getGravitationalParameter();
    const Orbit& orbit = _orbits.at(id);

    pos += orbit.getPosition(gravitationalParameter, time);

    return pos;
}

glm::dvec3 ParticleMap::getParticlePosition(const std::string& id, double time) const
{
    if (_orbits.find(id) == _orbits.end())
        return {0,0,0};

    const Particle* particle = &(_particles.at(id));
    glm::dvec3 pos(0,0,0);

    do
    {
        // TODO: this is shadowing a parameter with the same name
        const std::string& id = particle->getId();
        const std::string& parentId = _relationships.at(id);
        const Particle& parent = _particles.at(parentId);
        const double gravitationalParameter = parent.getGravitationalParameter();
        const Orbit& orbit = _orbits.at(id);

        pos += orbit.getPosition(gravitationalParameter, time);

        auto it = _relationships.find(parentId);
        if (it != _relationships.end())
        {
            particle = &parent;
        }
        else
        {
            particle = nullptr;
        }
    } while (particle != nullptr);

    return pos;
}

glm::dvec3 ParticleMap::getParticleLocalVelocity(const std::string& id, double time) const
{
    if (_orbits.find(id) == _orbits.end())
        return {0,0,0};

    const Particle* particle = &(_particles.at(id));
    glm::dvec3 vel(0,0,0);

    //const std::string& id = particle->getId();
    const std::string& parentId = _relationships.at(id);
    const Particle& parent = _particles.at(parentId);
    const double gravitationalParameter = parent.getGravitationalParameter();
    const Orbit& orbit = _orbits.at(id);

    vel += orbit.getVelocity(gravitationalParameter, time);

    return vel;
}

void ParticleMap::setParticleLocalVelocity(const std::string& id, const double& time, const glm::dvec3& velocity)
{
    // we want to convert a local position and velocity into a new set of keplerian elements.

    // first, get the local position of the e
    const Orbit& orbit = _orbits.at(id);
    const std::string& parentId = _relationships.at(id);
    const Particle& parent = _particles.at(parentId);
    const double gravitationalParameter = parent.getGravitationalParameter();
    const double meanAnomaly = orbit.getMeanAnomaly(gravitationalParameter, time);
    Vector6 cartesian(orbit.getCartesianCoordinates(meanAnomaly, gravitationalParameter));

    // update velocity, leave position as it is
    cartesian[astro::xVelocityIndex] = velocity.x;
    cartesian[astro::yVelocityIndex] = velocity.y;
    cartesian[astro::zVelocityIndex] = velocity.z;

    const Vector6 keplerian = astro::convertCartesianToKeplerianElements(cartesian, gravitationalParameter);

    Orbit newOrbit;
    newOrbit.setSemiMajorAxis(keplerian[astro::semiMajorAxisIndex]);
    newOrbit.setEccentricity(keplerian[astro::eccentricityIndex]);
    newOrbit.setInclination(keplerian[astro::inclinationIndex]);
    newOrbit.setArgumentOfPeriapsis(keplerian[astro::argumentOfPeriapsisIndex]);
    newOrbit.setLongitudeOfAscendingNode(keplerian[astro::longitudeOfAscendingNodeIndex]);

    setRelationship(parentId, id, newOrbit);
}

void ParticleMap::setParticle(const Particle& particle)
{
    _particles.insert({particle.getId(), particle});
}

void ParticleMap::setRelationship(const std::string& parentId, const std::string& childId, const Orbit& orbit)
{
    // map children to parent - children can only have one parent, so we use the map to
    // identify the parent directly using the child as the key
    _relationships.insert_or_assign(childId, parentId);
    _orbits.insert_or_assign(childId, orbit);
}
